from yta_multimedia.image.edition.filter.workers.sketch.filters import *
from yta_multimedia.image.edition.filter.workers.sketch.strokesort import *
from yta_multimedia.image.edition.filter.workers.sketch.util import *
from yta_general_utils.image.converter import ImageConverter
from random import *
from PIL import Image, ImageDraw, ImageOps

import  yta_multimedia.image.edition.filter.workers.sketch.perlin as perlin


no_cv = False
draw_contours = True
draw_hatch = True
show_bitmap = False
#resolution = 1024
resolution = 1920
hatch_size = 16
contour_simplify = 2

try:
    import numpy as np
    import cv2
except:
    no_cv = True

def find_edges(IM):
    if no_cv:
        #appmask(IM,[F_Blur])
        appmask(IM,[F_SobelX,F_SobelY])
    else:
        im = np.array(IM) 
        im = cv2.GaussianBlur(im,(3,3),0)
        im = cv2.Canny(im,100,200)
        IM = Image.fromarray(im)
    return IM.point(lambda p: p > 128 and 255)  


def getdots(IM):
    PX = IM.load()
    dots = []
    w,h = IM.size
    for y in range(h-1):
        row = []
        for x in range(1,w):
            if PX[x,y] == 255:
                if len(row) > 0:
                    if x-row[-1][0] == row[-1][-1]+1:
                        row[-1] = (row[-1][0],row[-1][-1]+1)
                    else:
                        row.append((x,0))
                else:
                    row.append((x,0))
        dots.append(row)
    return dots
    
def connectdots(dots):
    contours = []
    for y in range(len(dots)):
        for x,v in dots[y]:
            if v > -1:
                if y == 0:
                    contours.append([(x,y)])
                else:
                    closest = -1
                    cdist = 100
                    for x0,v0 in dots[y-1]:
                        if abs(x0-x) < cdist:
                            cdist = abs(x0-x)
                            closest = x0

                    if cdist > 3:
                        contours.append([(x,y)])
                    else:
                        found = 0
                        for i in range(len(contours)):
                            if contours[i][-1] == (closest,y-1):
                                contours[i].append((x,y,))
                                found = 1
                                break
                        if found == 0:
                            contours.append([(x,y)])
        for c in contours:
            if c[-1][1] < y-1 and len(c)<4:
                contours.remove(c)
    return contours


def getcontours(IM,sc=2):
    IM = find_edges(IM)
    IM1 = IM.copy()
    IM2 = IM.rotate(-90,expand=True).transpose(Image.FLIP_LEFT_RIGHT)
    dots1 = getdots(IM1)
    contours1 = connectdots(dots1)
    dots2 = getdots(IM2)
    contours2 = connectdots(dots2)

    for i in range(len(contours2)):
        contours2[i] = [(c[1],c[0]) for c in contours2[i]]    
    contours = contours1+contours2

    for i in range(len(contours)):
        for j in range(len(contours)):
            if len(contours[i]) > 0 and len(contours[j])>0:
                if distsum(contours[j][0],contours[i][-1]) < 8:
                    contours[i] = contours[i]+contours[j]
                    contours[j] = []

    for i in range(len(contours)):
        contours[i] = [contours[i][j] for j in range(0,len(contours[i]),8)]


    contours = [c for c in contours if len(c) > 1]

    for i in range(0,len(contours)):
        contours[i] = [(v[0]*sc,v[1]*sc) for v in contours[i]]

    for i in range(0,len(contours)):
        for j in range(0,len(contours[i])):
            contours[i][j] = int(contours[i][j][0]+10*perlin.noise(i*0.5,j*0.1,1)),int(contours[i][j][1]+10*perlin.noise(i*0.5,j*0.1,2))

    return contours


def hatch(IM,sc=16):
    PX = IM.load()
    w,h = IM.size
    lg1 = []
    lg2 = []
    for x0 in range(w):
        for y0 in range(h):
            x = x0*sc
            y = y0*sc
            if PX[x0,y0] > 144:
                pass
                
            elif PX[x0,y0] > 64:
                lg1.append([(x,y+sc/4),(x+sc,y+sc/4)])
            elif PX[x0,y0] > 16:
                lg1.append([(x,y+sc/4),(x+sc,y+sc/4)])
                lg2.append([(x+sc,y),(x,y+sc)])

            else:
                lg1.append([(x,y+sc/4),(x+sc,y+sc/4)])
                lg1.append([(x,y+sc/2+sc/4),(x+sc,y+sc/2+sc/4)])
                lg2.append([(x+sc,y),(x,y+sc)])

    lines = [lg1,lg2]
    for k in range(0,len(lines)):
        for i in range(0,len(lines[k])):
            for j in range(0,len(lines[k])):
                if lines[k][i] != [] and lines[k][j] != []:
                    if lines[k][i][-1] == lines[k][j][0]:
                        lines[k][i] = lines[k][i]+lines[k][j][1:]
                        lines[k][j] = []
        lines[k] = [l for l in lines[k] if len(l) > 0]
    lines = lines[0]+lines[1]

    for i in range(0, len(lines)):
        for j in range(0, len(lines[i])):
            lines[i][j] = int(lines[i][j][0]+sc*perlin.noise(i*0.5,j*0.1,1)),int(lines[i][j][1]+sc*perlin.noise(i*0.5,j*0.1,2))-j
    return lines


def apply_line_sketch_to_image(image: str, output_filename: str):
    # TODO: If this is for internal use we don't need 'output_filename'
    if not isinstance(image, np.ndarray) and not image:
        raise Exception('No "image" provided.')
    
    # TODO: Check 'output_filename'
    if isinstance(image, np.ndarray):
        IM = ImageConverter.numpy_image_to_pil(image)
    else:
        IM = Image.open(image)
    
    w, h = IM.size

    IM = IM.convert("L")
    IM = ImageOps.autocontrast(IM, 10)

    lines = []
    if draw_contours:
        lines += getcontours(IM.resize((resolution // contour_simplify, resolution // contour_simplify * h // w)), contour_simplify)
    if draw_hatch:
        lines += hatch(IM.resize((resolution // hatch_size, resolution // hatch_size * h // w)), hatch_size)

    lines = sortlines(lines)
    
    disp = Image.new("RGB", (resolution, resolution * h // w), (255, 255, 255))
    draw = ImageDraw.Draw(disp)
    for l in lines:
        draw.line(l, (0, 0, 0), 5)

    return disp

    disp.save(output_filename)

    # # We force it to be a svg file
    # output_filename = get_file_filename_without_extension(output_filename + '.svg')
    
    # f = open(output_filename, 'w')
    # f.write(makesvg(lines))
    # f.close()

    # TODO: Check svg to png conversion
    # https://stackoverflow.com/questions/6589358/convert-svg-to-png-in-python

    return output_filename


def makesvg(lines):
    out = '<svg xmlns="http://www.w3.org/2000/svg" version="1.1">'

    for l in lines:
        l = ",".join([str(p[0]*0.5)+","+str(p[1]*0.5) for p in l])
        out += '<polyline points="'+l+'" stroke="black" stroke-width="2" fill="none" />\n'
    out += '</svg>'

    return out



# if __name__ == "__main__":
#     parser = argparse.ArgumentParser(description='Convert image to vectorized line drawing for plotters.')
#     parser.add_argument('-i','--input',dest='input_path',
#         default='lenna',action='store',nargs='?',type=str,
#         help='Input path')

#     parser.add_argument('-b','--show_bitmap',dest='show_bitmap',
#         const = not show_bitmap,default= show_bitmap,action='store_const',
#         help="Display bitmap preview.")

#     parser.add_argument('-nc','--no_contour',dest='no_contour',
#         const = draw_contours,default= not draw_contours,action='store_const',
#         help="Don't draw contours.")
       
#     parser.add_argument('-nh','--no_hatch',dest='no_hatch',
#         const = draw_hatch,default= not draw_hatch,action='store_const',
#         help='Disable hatching.')

#     parser.add_argument('--no_cv',dest='no_cv',
#         const = not no_cv,default= no_cv,action='store_const',
#         help="Don't use openCV.")


#     parser.add_argument('--hatch_size',dest='hatch_size',
#         default=hatch_size,action='store',nargs='?',type=int,
#         help='Patch size of hatches. eg. 8, 16, 32')
#     parser.add_argument('--contour_simplify',dest='contour_simplify',
#         default=contour_simplify,action='store',nargs='?',type=int,
#         help='Level of contour simplification. eg. 1, 2, 3')

#     args = parser.parse_args()
    
#     draw_hatch = not args.no_hatch
#     draw_contours = not args.no_contour
#     hatch_size = args.hatch_size
#     contour_simplify = args.contour_simplify
#     show_bitmap = args.show_bitmap
#     no_cv = args.no_cv
#     sketch(args.input_path)


from yta_general_utils.image.parser import ImageParser
from yta_general_utils.programming.output import handle_output_filename
from yta_general_utils.file.enums import ImageFileExtension
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.parameter_validator import PythonValidator
from PIL import ImageDraw, Image
from typing import Union

import numpy as np
import cv2


class ImageMask:
    """
    Class to encapsulate and simplify the functionality
    related to image masking.
    """
    @staticmethod
    def apply_mask(image, mask_generator_method: 'function', output_filename: Union[str, None] = None):
        """
        Generate the provided 'mask_generator_method' (that
        must be one of the static methods of the MaskGenerator
        class) and apply it in the provided 'image' (that must
        be a valid image).

        This method return the result image as a ndim = 4 numpy
        array of not normalized values between 0 and 255, even
        the alpha channel.
        """
        # TODO: Is this the best way to handle a valid masking
        # generation method (?)
        # TODO: Maybe just check if the provided method is
        # contained in the MaskGenerator class (?)
        if not PythonValidator.is_class_staticmethod(MaskGenerator, mask_generator_method):
            raise Exception('The provided "mask_generator_method" parameter is not a valid MaskGenerator static method.')

        image = ImageParser.to_numpy(image, mode = 'RGB')

        # Obtain mask as ndim = 1 where black is transparent
        mask = ImageParser.to_numpy(mask_generator_method(image))[:, :, 1]
        
        # Add the mask to the Pillow Image
        image = np.dstack((image, mask))

        if output_filename:
            # TODO: Apply 'output' handler
            cv2.imwrite(output_filename, image)

        return image

        # TODO: How to write a numpy image
        #cv2.imwrite('atripa_trans.png', image_with_mask)
    
    

class MaskGenerator:
    """
    Class to encapsulate the different methods we have to
    dynamically generate image masks.
    """
    @staticmethod
    def rounded_corners(image, output_filename: Union[str, None] = None):
        """
        Generate a mask of the provided 'image' with the corners
        rounded and return it as a not normalized RGB Pillow Image 
        with all values between 0 (white) and 255 (black). This 
        means that the result, if converted to array, will be
        [0, 0, 0] for each white pixel and [255, 255, 255] for 
        black ones.

        Thank you: https://github.com/Zulko/moviepy/issues/2120#issue-2141195159
        """
        image = ImageParser.to_pillow(image)

        # Create a whole black image of the same size
        mask = Image.new('L', image.size, 'black')
        mask_drawing = ImageDraw.Draw(mask)

        # Generate the rounded corners mask
        w, h = image.size
        radius = 20
        # Rectangles to cover
        mask_drawing.rectangle([radius, 0, w - radius, h], fill = 'white')
        mask_drawing.rectangle([0, radius, w, h - radius], fill = 'white')
        # Circles at the corners: TL, TR, BL, BR
        mask_drawing.ellipse([0, 0, 2 * radius, 2 * radius], fill = 'white')
        mask_drawing.ellipse([w - 2 * radius, 0, w, 2 * radius], fill = 'white')
        mask_drawing.ellipse([0, h - 2 * radius, 2 * radius, h], fill = 'white')
        mask_drawing.ellipse([w - 2 * radius, h - 2 * radius, w, h], fill = 'white')

        if output_filename:
            output_filename = handle_output_filename(output_filename, ImageFileExtension.PNG)
            mask.save(output_filename)

        return mask

    def ticket(image, output_filename: Union[str, None] = None):
        image = ImageParser.to_pillow(image)

        # Create a whole black image of the same size
        mask = Image.new('L', image.size, 'white')
        mask_drawing = ImageDraw.Draw(mask)

        # Generate the mask
        w, h = image.size
        triangle_size = 15

        for i in range(0, w, triangle_size):
            mask_drawing.polygon([(i, h), (i + triangle_size, h), (i + triangle_size // 2, h - triangle_size)], fill = 'black')
            mask_drawing.polygon([(i, 0), (i + triangle_size, 0), (i + triangle_size // 2, triangle_size)], fill = 'black')

        if output_filename:
            output_filename = handle_output_filename(output_filename, ImageFileExtension.PNG)
            mask.save(output_filename)

        return mask

    # TODO: Add more effects 



"""
    Thanks to https://omaraflak.medium.com/b%C3%A9zier-interpolation-8033e9a262c2
    for interpolation explanation and code.

    Thanks to https://github.com/vmichals/python-algos/blob/master/catmull_rom_spline.py
    for the catmull-rom-spline interpolation code.
"""
from yta_general_utils.programming.parameter_validator import NumberValidator, PythonValidator
from yta_general_utils.image.region import Coordinate
from typing import Callable
from math import comb

import random
import bezier
import numpy as np
import matplotlib.pyplot as plt


class NormalizedCoordinate(Coordinate):
    """
    A coordinate to be used in the graphic rate
    function building system, which is a normalized
    2D environment with values between 0.0 and 1.0.
    """
    def __init__(self, x: float, y: float):
        if not NumberValidator.is_number_between(x, 0.0, 1.0) or not NumberValidator.is_number_between(y, 0.0, 1.0):
            raise Exception(f'The provided "x" and/or "y" parameters {str(x)},{str(y)} values are not between 0.0 and 1.0.')
        
        super().__init__(x, y)

    @staticmethod
    def generate(amount: int = 1):
        """
        Generate 'amount' normalized coordinates with random
        values that are returned as an array of instances.

        The amount parameter is limited to the interval 
        [1, 100].
        """
        if not NumberValidator.is_number_between(amount, 1, 100):
            raise Exception(f'The provided "amount" parameter "{str(amount)}" is not a number between 1 and 100.')
        
        return [NormalizedCoordinate(random.uniform(0, 1), random.uniform(0, 1)) for _ in range(amount)]
    
    @staticmethod
    def to_numpy(coordinates: list):
        """
        Convert a list of BizierCoordinate 'coordinates' to
        numpy array to be able to work with them.

        This method does the next operation:
        np.array([[coord.x, coord.y] for coord in coordinates])
        """
        if not PythonValidator.is_list(coordinates):
            if not PythonValidator.is_instance(coordinates, NormalizedCoordinate):
                raise Exception('The provided "coordinates" parameter is not a list of NormalizedCoordinates nor a single NormalizedCoordinate instance.')
            else:
                coordinates = [coordinates]
        elif any(not PythonValidator.is_instance(coordinate, NormalizedCoordinate) for coordinate in coordinates):
            raise Exception('At least one of the provided "coordinates" is not a NormalizedCoordinate instance.')

        return np.array([coordinate.to_array() for coordinate in coordinates])


# Here below are some functions to obtain the graphic
# between points
class InterpolationFormula:
    """
    Class to encapsulate interpolation formulas.
    """
    @staticmethod
    def catmull_rom_spline(spline_x: float, p0, p1, p2, p3):
        """
        Function to evaluate one spline of the Catmull-Rom
        between two points by using 4 control points.

        It returns the x,y point in the spline.

        Computes interpolated y-coord for given x-coord using
        Catmull-Rom. The 'spline_x' is the current position in
        the graphic interpolation between the points p0 and p1
        represented by a value between 0 and 1.

        Computes an interpolated y-coordinate for the given
        x-coordinate between the support points v1 and v2. The
        neighboring support points v0 and v3 are used by
        Catmull-Rom to ensure a smooth transition between the
        spline segments.
        """
        c1 = 1. * p1
        c2 = -.5 * p0 + .5 * p2
        c3 = 1. * p0 + -2.5 * p1 + 2. * p2 -.5 * p3
        c4 = -.5 * p0 + 1.5 * p1 + -1.5 * p2 + .5 * p3

        return (((c4 * spline_x + c3) * spline_x + c2) * spline_x + c1)
    
    @staticmethod
    def linear(spline_x: float, p0, p1):
        """
        Function to evaluate one spline of a linear function
        between two points.

        It returns the x,y point in the spline.

        Computes interpolated y-coord for given x-coord using
        linear function.
        """
        x = p0[0] + (p1[0] - p0[0]) * spline_x
        y = p0[1] + (p1[1] - p0[1]) * spline_x

        return (x, y)

class GraphicRateFunction:
    """
    Class to simplify the way we work with rate functions
    being able to personalize a graphic that will allow us
    using it as a rate function to create our own effects.

    We receive a list of points that are coordinates of the
    graph we want to create as a rate function, and the 
    code will analyze it and obtain the function that allows
    us to determine the value of any 't' value between 0.0
    and 1.0.
    
    The 't' value provided is representing the distance in the
    graphic curve that has been generated with the points
    provided and the interpolation method applied. This curve
    is representing some effect, animation or transition
    duration, from 0 to 1. So, if the provided 't' is 0.20,
    that is representing the moment in which the 20% of the
    animation has been played, and will return the 'y' value
    that must be used in the animation for that specific 
    moment.

    The points provided must be consecutive. The 'x' atribute
    of each point must be greater than the previous one.

    TODO: This class is using the catmull-rom interpolation
    algorithm, but its name is generic, so I should adapt
    this class to accept different interpolation methods to
    be used between the provided coordinates (points).
    """
    points: np.ndarray
    """
    The list of points [x, y] the graph will go through.
    """
    interpolation_formula: Callable
    """
    The method to calculate the interpolation between each
    pair of points.

    TODO: By now it is the same for the whole graph but it
    could change in a near future and be specific for each
    pair of points.
    """

    def __init__(self, points: list[NormalizedCoordinate], interpolation_formula: Callable):
        if not PythonValidator.is_list(points) or len(points) < 4 or any(not PythonValidator.is_instance(point, NormalizedCoordinate) for point in points):
            raise Exception(f'The "points" parameter must be a list of NormalizedCoordinates with a minimum of 4 elements.')
    
        if not PythonValidator.is_class_staticmethod(InterpolationFormula, interpolation_formula):
            raise Exception('The provided "interpolation_formula" is not a valid InterpolationFormula class method.')

        # TODO: Maybe if less than 4 elements: add 2 in between those 2
        if not all(points[i].x < points[i + 1].x for i in range(len(points) - 1)):
            raise Exception('Points must be consecutive, so the "x" value of each point must be greater than the previous one. Please, check the provided "points" parameter.')

        self.points = np.array([[coord[0], coord[1]] for coord in NormalizedCoordinate.to_numpy(points)])
        self.interpolation_formula = interpolation_formula

    def get_coord(self, x: float):
        """
        Evaluate the function for the provided 'x' value and
        return the graphic coordinate corresponding to that
        'x' value, that must be a normalized value (between
        0 and 1, both inclusive) that represents the distance
        in the generated graphic.

        A graphic with 5 coordinates will generate 4 splines
        of the same length. The start of each of those splines
        will be [x = 0, x = 0.25, x = 0.5, x = 0.75]. So, if
        you provide x = 0.20, the returned coordinate will be
        contained in the first spline because it is below 0.25.

        TODO: This method is using the 'catmull_rom_spline'
        interpolation, so I should adapt this method to accept
        an 'interpolation_method' parameter and calculate 
        everything according to that specific method.
        """
        if not NumberValidator.is_number_between(x, 0.0, 1.0):
            raise Exception('The provided "x" parameter "{str(x)}" must be  between 0 and 1, both inclusive.')

        num_of_splines = len(self.points) - 1
        spline_size = 1 / num_of_splines

        if x == 1.0:
            spline_index = num_of_splines - 1
        else:
            # Check which spline we need to use
            spline_index = int(x // spline_size)

        # If we receive 0.28 => 0.28 - 0.25 = 0.03 / spline_size
        # A distance of 0.03 in the global graph is lower than the
        # real distance in any splice
        # 0.28 = 0.28 % (1 * 0.25) / 0.25 = 0.03 / 0.25 = 0.12
        previous_spline_size = spline_index * spline_size

        if previous_spline_size:
            x = x % (spline_index * spline_size) / spline_size
        else:
            x = x / spline_size

        if PythonValidator.is_class_staticmethod(InterpolationFormula, self.interpolation_formula, InterpolationFormula.catmull_rom_spline.__name__):
            if spline_index == 0: # First spline
                # We need to estimate the first point
                p0 = self.points[0] - (self.points[1] - self.points[0])
                p1 = self.points[0]
                p2 = self.points[1]
                p3 = self.points[2]
            elif spline_index == (num_of_splines - 1): # Last spline
                p0 = self.points[spline_index - 1]
                p1 = self.points[spline_index]
                p2 = self.points[spline_index + 1]
                # We need to estimate the last point
                p3 = self.points[spline_index + 1] + (self.points[spline_index + 1] - self.points[spline_index])
            else:
                p0 = self.points[spline_index - 1]
                p1 = self.points[spline_index]
                p2 = self.points[spline_index + 1]
                p3 = self.points[spline_index + 2]

            coord = self.interpolation_formula(x, p0, p1, p2, p3)
        elif PythonValidator.is_class_staticmethod(InterpolationFormula, self.interpolation_formula, InterpolationFormula.linear.__name__):
            p0 = self.points[spline_index]
            p1 = self.points[spline_index + 1]

            coord = self.interpolation_formula(x, p0, p1)
        
        # We want the 'y' value as we are representing
        # something in a graphic, but we get the whole
        # coord at this point
        return coord
    
    def get_y(self, x: float):
        """
        Evaluate the function for the provided 'x' value and
        return the graphic coordinate 'y' value corresponding
        to that 'x' value that represents the distance in the
        generated graphic.

        A graphic with 5 coordinates will generate 4 splines
        of the same length. The start of each of those splines
        will be [x = 0, x = 0.25, x = 0.5, x = 0.75]. So, if
        you provide x = 0.20, the returned coordinate will be
        contained in the first spline because it is below 0.25.
        """
        return self.get_coord(x)[1]

    def plot(self):
        """
        This method is just for testing and shows the graphic
        using the matplotlib library.
        """
        graph = np.array([self.get_coord(t) for t in np.linspace(0, 1, len(self.points) * 100)])

        plt.plot(graph[:, 0], graph[:, 1], label = 'Graphic representation', color = 'blue')
        plt.scatter(self.points[:, 0], self.points[:, 1], color = 'red', label = 'Control points')
        plt.legend()
        plt.title('Graphic representation')
        plt.xlabel('x')
        plt.ylabel('y')
        plt.grid(True)
        plt.show()

    # TODO: If we want to apply one GraphicRateFunction to some
    # of our moviepy animations, we will be working with a 't'
    # value representing the time of the video, that can easily
    # be greater than 1, so we need to normalize it with the
    # video duration and use that normalized value to provide it
    # to this 'get_y()' or 'get_coord()' method.



from yta_general_utils.math.rate_functions import RateFunction





# TODO: I keep this code below because it was almost working
# but need refactor to be adapted to the way I want to use
# this interpolation functionality. I need to join pairs of
# points to build a Graph concept to be able to use it as the
# rate function definer
class BezierCurve:
    """
    Class to encapsulate and simplify the funcitonality
    related to bezier curves.
    """
    curve: bezier.Curve
    """
    The instance of the Curve from the bezier library.
    """

    def __init__(self, control_nodes: list[NormalizedCoordinate]):
        if not PythonValidator.is_list(control_nodes):
            if not PythonValidator.is_instance(control_nodes, NormalizedCoordinate):
                raise Exception('The provided "control_nodes" parameter is not a list of NormalizedCoordinate instances nor a single NormalizedCoordinate instance.')
            else:
                control_nodes = [control_nodes]

        if any(not PythonValidator.is_instance(control_node, NormalizedCoordinate) for control_node in control_nodes):
            raise Exception(f'At least one of the provided "control_nodes" parameter is not a NormalizedCoordinate instance.')
        
        nodes_np_array = [[control_node.x for control_node in control_nodes], [control_node.y for control_node in control_nodes]]
        degree = len(control_nodes) - 1

        self.curve = bezier.Curve(nodes_np_array, degree = degree)
        # TODO: Anything else (?)

    def get_curve_point_value(self, t: float):
        """
        Obtain the curve value for the provided 't' that must
        be a value between 0.0 and 1.0.
        """
        if not NumberValidator.is_number_between(t, 0.0, 1.0):
            raise Exception(f'The provided "t" parameter value "{str(t)}" is not a valid value. Must be between 0.0 and 1.0.')

        return self.curve.evaluate(t)

    def plot(self, do_show_control_nodes: bool = True):
        """
        Plot and show the bezier curve using matplotlib.
        """
        # Set 100 points of the curve to draw it
        curve_points = self.curve.evaluate_multi(np.linspace(0, 1, 100))
        nodes = self.curve.nodes

        plt.plot(curve_points[0], curve_points[1], label = "Bezier curve", color = 'b')

        if do_show_control_nodes:
            plt.scatter(nodes[0], nodes[1], color = 'r', label = "Control nodes")
            # Draw control nodes lines in-between
            for i in range(len(nodes[0]) - 1):
                plt.plot([nodes[0][i], nodes[0][i + 1]], [nodes[1][i], nodes[1][i + 1]], 'r--')

        plt.legend()
        plt.title(f'Bezier curve (degree = {str(len(nodes[0]) - 1)})')
        plt.grid(True)
        plt.axis('equal')
        plt.show()

    @staticmethod
    def generate_random():
        # TODO: This is the interpolation process
        points = NormalizedCoordinate.generate(8)
        points.insert(0, NormalizedCoordinate(0, 0))
        points.append(NormalizedCoordinate(1, 1))
        path = evaluate_bezier(points, 50)

        #points = np.array([point.to_array() for point in points])
        points = NormalizedCoordinate.to_numpy(points)
        # extract x & y coordinates of points
        x, y = points[:,0], points[:,1]
        px, py = path[:,0], path[:,1]

        # plot
        plt.figure(figsize=(11, 8))
        plt.plot(px, py, 'b-')
        plt.plot(x, y, 'ro')
        plt.show()

        return
        # TODO: I'm working with these below to be able to build
        # the BezierCurve instance by providing the path cooridnates
        # from which we interpolate and calculate the curve

        # A and B are the 2 control nodes for each 2 consecutive
        # points (it is a cubic interpolation)
        first_control_nodes, second_control_nodes = get_bezier_coeficient(points)

        curve_path = [
            get_cubic_bezier_formula(points[i], first_control_nodes[i], second_control_nodes[i], points[i + 1])
            for i in range(len(points) - 1)
        ]

        # TODO: This is to print the whole curve
        n = 100
        real_curve_path_with_100_points = np.array([fun(t) for fun in curve_path for t in np.linspace(0, 1, n)])

    @staticmethod
    def get_bezier_curve(points: list[NormalizedCoordinate]):
        points = NormalizedCoordinate.to_numpy(points)
        first_control_nodes, second_control_nodes = get_bezier_coeficient(points)

        return BezierCurve()
    
    @staticmethod
    def get_bezier_curve_from_points(points: list[NormalizedCoordinate]):
        # I obtain the control nodes (2, as cubic bezier function)
        # for each point
        first_control_nodes, second_control_nodes = get_bezier_coeficient(points)
        bezier_function = [
            get_cubic_bezier_formula(points[i], first_control_nodes[i], second_control_nodes[i], points[i + 1])
            for i in range(len(points) - 1)
        ]

        # I have many bezier curves concatenated

        # This generates 100 points for each bezier curve
        np.array([fun(t) for fun in bezier_function for t in np.linspace(0, 1, n)])

def get_bezier_coeficient(points):
    """
    Find the A and B points. These are the control nodes for
    each bezier cubic curve point, as it is using two control
    nodes (A and B) that are the ones we are looking for.

    TODO: Is this ok (?)

    TODO: How to apply this to a non-cubic bezier curve (?)
    """
    # since the formulas work given that we have n+1 points
    # then n must be this:
    n = len(points) - 1

    # Build the coefficents matrix
    C = 4 * np.identity(n)
    np.fill_diagonal(C[1:], 1)
    np.fill_diagonal(C[:, 1:], 1)
    C[0, 0] = 2
    C[n - 1, n - 1] = 7
    C[n - 1, n - 2] = 2

    # Build the points vector
    P = [2 * (2 * points[i] + points[i + 1]) for i in range(n)]
    P[0] = points[0] + 2 * points[1]
    P[n - 1] = 8 * points[n - 1] + points[n]

    # Solve the system and find control nodes A and B
    A = np.linalg.solve(C, P)
    B = [0] * n
    for i in range(n - 1):
        B[i] = 2 * points[i + 1] - A[i + 1]
    B[n - 1] = (A[n - 1] + points[n]) / 2

    return A, B

# TODO: This below has not been checked
def get_lineal_bezier_formula(node1, node2):
    return lambda t: (1 - t) * np.array(node1) + t * np.array(node2)

# TODO: This below has not been checked
def get_cuadratic_bezier_formula(node1, node2, node3):
    return lambda t: (1 - t)**2 * np.array(node1) + 2 * (1 - t) * t * np.array(node2) + t**2 * np.array(node3)

# TODO: This is the only one that is working for sure
def get_cubic_bezier_formula(node1, node2, node3, node4):
    """
    Return the general Bezier cubic formula given 4 control
    nodes.
    """
    return lambda t: np.power(1 - t, 3) * node1 + 3 * np.power(1 - t, 2) * t * node2 + 3 * (1 - t) * np.power(t, 2) * node3 + np.power(t, 3) * node4

# TODO: This below has not been checked
def get_n_bezier_formula(nodes: list):
    def x(t):
        N = len(nodes)
        B_t = np.zeros_like(nodes[0])  # Inicializar el punto en 0 (misma dimensión que los puntos de control)
        
        for i in range(N):
            # Coeficiente binomial
            coef_binom = comb(N - 1, i)
            # Fórmula de Bézier general
            B_t += coef_binom * (1 - t)**(N - 1 - i) * t**i * np.array(nodes[i])

    return lambda t: x(t)

# return one cubic curve for each consecutive points
def get_bezier_cubic(points):
    """
    Return a cubir curve for each consecutive points.
    """
    A, B = get_bezier_coeficient(points)

    return [
        get_cubic_bezier_formula(points[i], A[i], B[i], points[i + 1])
        for i in range(len(points) - 1)
    ]

def evaluate_bezier(points: list[NormalizedCoordinate], n):
    """
    Evaluate each cubic curve on the range [0, 1]
    sliced in n points.

    TODO: Maybe rename to interpolate (?)
    """
    #points = [point.to_array() for point in points]
    # Transform BezierCoordinates to array
    points = NormalizedCoordinate.to_numpy(points)
    #points = np.array([point.to_array() for point in points])

    return np.array([fun(t) for fun in get_bezier_cubic(points) for t in np.linspace(0, 1, n)])



# TODO: This code below was working
"""
    points = BezierControlNode.generate(8)
    points.insert(0, BezierControlNode(0, 0))
    points.append(BezierControlNode(1, 1))
    path = evaluate_bezier(points, 50)


    points = np.array([point.to_array() for point in points])
    # extract x & y coordinates of points
    x, y = points[:,0], points[:,1]
    px, py = path[:,0], path[:,1]

    # plot
    plt.figure(figsize=(11, 8))
    plt.plot(px, py, 'b-')
    plt.plot(x, y, 'ro')
    plt.show()
"""



"""
        This below is a remaning that I want to check only
        and then, when everything is clear, remove it.
"""
# """
# I copied all these files from the "software" project in which
# I initially created them, just to preserve and adapt the ones
# we weant to keep in the code.
# """
# from yta_multimedia.video.edition.effect.moviepy.testing.position_effect import PositionEffect
# from moviepy.editor import CompositeVideoClip, ColorClip

# def clip_with_effect(clip, effect: PositionEffect, **kwargs):
#     """
#     This method will create a black background clip and will put
#     as an overlay the provided 'clip' with the also provided
#     'effect' applied on it.

#     This method will use the kwargs['time'] as the effect time if provided,
#     or will use the 'destination_clip' duration if no time provided. This
#     method will also crop the 'time' if provided to, as maximum, the
#     'destination_clip' duration.

#     Pay atenttion to the parameters of the PositionEffect you are trying
#     to use.

#     This method will return a CompositeVideoClip that includes the 
#     'destination_clip' and the effect applied over it.
#     """
#     background_clip = ColorClip(clip.size, [0, 0, 0], duration = clip.duration)

#     return __add_clip_with_effect(clip, effect = effect, destination_clip = background_clip, **kwargs)

# def add_clip_with_effect(clip, effect: PositionEffect, destination_clip, **kwargs):
#     """
#     Adds the provided 'clip' with the also provided 'effect' over the
#     'destination_clip'. The provided 'clip' is the clip that will be used 
#     in the effect. For example, an ImageClip that contains an emoji that 
#     is going to be slided in and out.

#     This method will use the kwargs['time'] as the effect time if provided,
#     or will use the 'destination_clip' duration if no time provided. This
#     method will also crop the 'time' if provided to, as maximum, the
#     'destination_clip' duration.

#     Pay atenttion to the parameters of the PositionEffect you are trying
#     to use.

#     This method will return a CompositeVideoClip that includes the 
#     'destination_clip' and the effect applied over it.
#     """
#     return __add_clip_with_effect(clip, effect = effect, destination_clip = destination_clip, **kwargs)

# def __add_clip_with_effect(clip, effect: PositionEffect, destination_clip, **kwargs):
#     """
#     This method is to avoid duplicating the code. Only for internal use.
#     """
#     if not 'time' in kwargs:
#         kwargs['time'] = destination_clip.duration

#     if kwargs.get('time') > destination_clip.duration:
#         kwargs['time'] = destination_clip.duration

#     if kwargs.get('time') < clip.duration:
#         return CompositeVideoClip([
#             destination_clip,
#             *effect(clip.subclip(0, kwargs['time']), **kwargs),
#             clip.subclip(kwargs['time'], clip.duration).set_start(kwargs['time'])
#         ])

#     return CompositeVideoClip([
#         destination_clip,
#         *effect(clip, **kwargs)
#     ])


from yta_multimedia.alphascreen.image_alphascreen import ImageAlphascreen
from yta_multimedia.alphascreen.video_alphascreen import VideoAlphascreen
from yta_multimedia.resources import Resource
from yta_general_utils.checker.url import url_is_ok
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.checker.url import is_google_drive_url
from yta_general_utils.downloader.google_drive import GoogleDriveResource
from yta_multimedia.alphascreen.consts import ALPHASCREENS_FOLDER


class Alphascreen:
    """
    This class is an initializer of the ImageAlphascreen and 
    VideoAlphascreen classes so you provide the valid filename
    or Google Drive url of an alphascreen to the 'init' method
    and you get the corresponding ImageAlphascreen or 
    VideoAlphascreen initialized and ready to work.

    If you provide a Google Drive url the system will identify
    the resource and download it only if necessary, as it will
    create locally a unique identifier with the Google Drive ID
    so once it's been downloaded the first time it will be get
    from the local source the next time.

    This will not be possible if a local path is given as the
    system will not be able to identify the resource and the
    alphascreen processing time will be long in any execution.
    """
    @classmethod
    def init(cls, filename_or_google_drive_url: str):
        """
        Parses the provided parameter 'filename_or_google_drive_url' and
        checks if it is a valid filename or a valid Google Drive url. If
        valid, it dowloads the resource (if necessary) and checks if the
        file is an image or a video and returns the corresponding
        ImageAlphascreen or VideoAlphascreen.

        This method will raise an Exception if something goes wrong.
        """
        is_a_file = FileValidator.is_file(filename_or_google_drive_url)
        is_url = False
        filename = filename_or_google_drive_url

        if not is_a_file:
            is_url = url_is_ok(filename_or_google_drive_url)

            if not is_url:
                raise Exception(f'The provided alphascreen "{filename_or_google_drive_url}" parameter is not a file nor a valid url.')
            
            if not is_google_drive_url(filename_or_google_drive_url):
                raise Exception(f'The provided alphascreen "url" parameter "{filename_or_google_drive_url}" is not a valid Google Drive url.')
            
            google_drive_id = GoogleDriveResource(filename_or_google_drive_url).id
            folder = ALPHASCREENS_FOLDER + google_drive_id + '/'
            # Extension will be replaced with real one
            filename = Resource.get(filename_or_google_drive_url, f'{folder}alphascreen.png')

        # We have a valid filename, lets check if video or image
        if FileValidator.file_is_image_file(filename):
            return ImageAlphascreen(filename)
        elif FileValidator.file_is_video_file(filename):
            return VideoAlphascreen(filename)
        else:
            raise Exception(f'The alphascreen file "{filename}" is not a valid image or video file.')
    


from yta_multimedia.video.edition.effect.open_close.fade_in_video_effect import FadeInVideoEffect
from yta_multimedia.video.edition.effect.open_close.fade_out_video_effect import FadeOutVideoEffect
from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, concatenate_videoclips, VideoClip, ColorClip
from typing import Union


class BlinkVideoEffect(VideoEffect):
    """
    This method makes the provided video blink, that is a composition of
    a FadeOut and a FadeIn consecutively to build this effect. The duration
    will be the whole clip duration. The FadeIn will last the half of the
    clip duration and the FadeOut the other half.

    The 'color' parameter is the color you want for the blink effect as the
    background color. The default value is black ([0, 0, 0]).
    """

    @staticmethod
    def apply(clip: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], color = None):
        """
        Applies the effect to the provided 'clip'.
        """
        clip = VideoEffect.parse_moviepy_video(clip)

        if color is None:
            color = [0, 0, 0]

        half_duration = clip.duration / 2
        clip = concatenate_videoclips([
            FadeOutVideoEffect.apply(clip.subclip(0, half_duration), duration = half_duration, color = color),
            FadeInVideoEffect.apply(clip.subclip(half_duration, clip.duration), duration = half_duration, color = color)
        ])

        return clip


from yta_multimedia.image.greenscreen import remove_greenscreen_from_image
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.utils.ffmpeg_handler import FfmpegHandler, FfmpegPixelFormat
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.temp import create_custom_temp_filename
from typing import Union
from moviepy.editor import VideoFileClip


def remove_greenscreen_from_video(video_input: Union[VideoFileClip, str], output_filename: str = None):
    """
    Removes the green screen from the 'video_input' video so
    you get a final 'output_filename' video with transparent
    layer.
    """
    video = VideoParser.to_moviepy(video)
    
    if PythonValidator.is_string(video_input):
        if not FileValidator.file_is_video_file(video_input):
            return None
        
        video_input = VideoFileClip(video_input)

    # TODO: I think I need to write the frame filenames in a file to be used
    # with the concat flag and ffmpeg library

    # Export all frames
    original_frames_array = []
    for frame in video_input.iter_frames():
        frame_name = create_custom_temp_filename('tmp_frame_' + str(len(original_frames_array)) + '.png')
        original_frames_array.append(frame_name)
    video_input.write_images_sequence(create_custom_temp_filename('tmp_frame_%01d.png'), logger = 'bar')

    # Remove green screen of each frame and store it
    processed_frames_array = []
    for index, frame in enumerate(original_frames_array):
        tmp_frame_filename = create_custom_temp_filename('tmp_frame_processed_' + str(index) + '.png')
        processed_frames_array.append(tmp_frame_filename)
        remove_greenscreen_from_image(frame, tmp_frame_filename)

    # Rebuild the video
    if not output_filename:
        output_filename = create_custom_temp_filename('greenscreen_video_removed_background.mov')
    # TODO: What if 'output_filename' is not .mov but it has transparency,
    # maybe we should force '.mov'

    video, _ = FfmpegHandler.concatenate_images(processed_frames_array, frame_rate = 30, pixel_format = FfmpegPixelFormat.YUV420p, output_filename = output_filename)

    return video

    #parameters = ['ffmpeg', '-y', '-i', create_custom_temp_filename('tmp_frame_processed_%01d.png'), '-r', '30', '-pix_fmt', 'yuva420p', output_filename]
    #run(parameters)
        
    # https://stackoverflow.com/a/77608713
    #ImageSequenceClip(processed_frames_array, fps = clip.fps).set_audio(clip.audio).write_videofile(output_filename, codec = 'hap_alpha', ffmpeg_params = ['-c:v', 'hap', '-format', 'hap_alpha', '-vf', 'chromakey=black:0.1:0.1'])
    # https://superuser.com/questions/1779201/combine-pngs-images-with-transparency-into-a-video-and-merge-it-on-a-static-imag
    # ffmpeg -y -i src/tmp/%d.png -c:v libx264 -vf fps=25 -pix_fmt yuva420p land.mov
    
    #clip = ImageSequenceClip(processed_frames_array, fps = clip.fps).set_audio(clip.audio).write_videofile(output_filename, codec = 'libx264', audio_codec = 'aac', temp_audiofile = 'temp-audio.m4a', remove_temp = True)


from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.image.parser import ImageParser

import numpy as np


class NumpyFrameHelper:
    """
    Class to encapsulate functionality related to numpy
    frames.
    """
    @staticmethod
    def is_rgb_not_normalized(frame: np.ndarray):
        """
        Check if the provided 'frame' is a numpy array of
        ndim = 3, dtype = np.uint8 and all the values (3)
        are between 0 and 255.
        """
        frame = ImageParser.to_numpy(frame)

        if not PythonValidator.is_numpy_array(frame):
            raise Exception('The provided "frame" parameter is not a numpy array.')
        
        return frame.ndim == 3 and frame.dtype == np.uint8 and frame.shape[2] == 3 and np.all((frame >= 0) & (frame <= 255))
    
    @staticmethod
    def is_rgb_normalized(frame: np.ndarray):
        """
        Check if the provided 'frame' is a numpy array of
        ndim = 3, dtype = np.float64|np.float32 and all 
        the values (3) are between 0.0 and 1.0.
        """
        frame = ImageParser.to_numpy(frame)

        if not PythonValidator.is_numpy_array(frame):
            raise Exception('The provided "frame" parameter is not a numpy array.')
        
        return frame.ndim == 3 and frame.dtype in (np.float64, np.float32) and frame.shape[2] == 3 and np.all((frame >= 0.0) & (frame <= 1.0))
    
    @staticmethod
    def is_rgba_not_normalized(frame: np.ndarray):
        """
        Check if the provided 'frame' is a numpy array of
        ndim = 3, dtype = np.uint8 and all the values (4)
        are between 0 and 255.
        """
        """
        Check if the provided 'frame' is a numpy array of
        ndim = 3, dtype = np.uint8 and all the values (3)
        are between 0 and 255.
        """
        frame = ImageParser.to_numpy(frame)

        if not PythonValidator.is_numpy_array(frame):
            raise Exception('The provided "frame" parameter is not a numpy array.')
        
        return frame.ndim == 3 and frame.dtype == np.uint8 and frame.shape[2] == 4 and np.all((frame >= 0) & (frame <= 255))
    
    @staticmethod
    def is_rgba_normalized(frame: np.ndarray):
        """
        Check if the provided 'frame' is a numpy array of
        ndim = 3, dtype = np.float64|np.float32 and all 
        the values (4) are between 0.0 and 1.0.
        """
        frame = ImageParser.to_numpy(frame)

        if not PythonValidator.is_numpy_array(frame):
            raise Exception('The provided "frame" parameter is not a numpy array.')
        
        return frame.ndim == 3 and frame.dtype in (np.float64, np.float32) and frame.shape[2] == 4 and np.all((frame >= 0.0) & (frame <= 1.0))
    
    @staticmethod
    def is_alpha_not_normalized(frame: np.ndarray):
        """
        Check if the provided 'frame' is a numpy array of
        ndim = 2, dtype = np.uint8 and all
        the values are between 0 and 255.
        """
        frame = ImageParser.to_numpy(frame)

        if not PythonValidator.is_numpy_array(frame):
            raise Exception('The provided "frame" parameter is not a numpy array.')
        
        return frame.ndim == 2 and frame.dtype == np.uint8 and np.all((frame >= 0) & (frame <= 255))

    @staticmethod
    def is_alpha_normalized(frame: np.ndarray):
        """
        Check if the provided 'frame' is a numpy array of
        ndim = 2, dtype = np.float64|np.float32 and all
        the values are between 0.0 and 1.0.
        """
        frame = ImageParser.to_numpy(frame)

        if not PythonValidator.is_numpy_array(frame):
            raise Exception('The provided "frame" parameter is not a numpy array.')
        
        return frame.ndim == 2 and frame.dtype in (np.float64, np.float32) and np.all((frame >= 0.0) & (frame <= 1.0))





from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames.video_frame_extractor import VideoFrameExtractor
from yta_multimedia.video.generation import generate_video_from_image
from yta_multimedia.video.edition.effect.speed.change_speed_effect import ChangeSpeedVideoEffect
from yta_general_utils.programming.enum import YTAEnum as Enum
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, concatenate_videoclips
from typing import Union


# TODO: Move this to a better place
class ExtendVideoMode(Enum):
    """
    This is a Enum to set the parameter option to extend the video
    duration with one of these modes (strategies).
    """
    LOOP = 'loop'
    """
    This mode will make the video loop (restart from the begining)
    until it reaches the expected duration.
    """
    FREEZE_LAST_FRAME = 'freeze_last_frame'
    """
    This mode will freeze the last frame of the video and extend 
    it until it reaches the expected duration.
    """
    SLOW_DOWN = 'slow_down'
    """
    This mode will change the speed of the provided video to make
    it fit the needed duration by deccelerating it. As you should
    know, this method changes the whole video duration so the 
    result could be unexpected. Use it carefully.
    """
    SPEED_UP_OR_SLOW_DOWN = 'speed_up_or_slow_down'
    """
    This mode will change the speed of the provided video to make
    it fit the needed duration by accelerating or deccelerating 
    it. This will speed up the video if the requested duration is
    shorter than the provided 'video', or will slow it down if it
    is larger. The result could be not as good as expected because
    of the speed change. Use it carefully.
    """
    # TODO: This was thought to enlarge the video if needed and 
    # only subclip it if 'duration' was less than the real video
    # duration, but now we are able to speed it up instead of 
    # subclipping the video. Maybe we need 'enshort_mode' and
    # 'enlarge_mode' instead of only 'mode' parameter to be able
    # to chose both strategies. This will simplify the if logic
    # and make it more customizable but being solid.


def set_video_duration(video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip], duration = float, mode: ExtendVideoMode = ExtendVideoMode.LOOP):
    """
    This method will return a copy of the provided 'video' with the desired
    'duration' by applying crops or loops. If the provided 'duration' is
    lower than the actual 'video' duration, it will be shortened. If it is
    greater, it will be looped until we reach the desired 'duration'.

    The 'mode' provided will determine the way in which we extend the video
    duration if needed.

    This method makes a 'video.copy()' internally to work and avoid problems.
    """
    video = VideoParser.to_moviepy(video)

    if not duration:
        raise Exception('No "duration" provided.')
    
    if not mode:
        mode = ExtendVideoMode.LOOP
    else:
        mode = ExtendVideoMode.to_enum(mode)

    final_video = video.copy()

    if video.duration > duration:
        # TODO: Take a look at the TODO on the top, please
        if mode == ExtendVideoMode.SPEED_UP_OR_SLOW_DOWN:
            final_video = ChangeSpeedVideoEffect.apply(final_video, duration)
        else:
            final_video = final_video.subclip(0, duration)
    elif video.duration < duration:
        if mode == ExtendVideoMode.LOOP:
            times_to_loop = (int) (duration / video.duration) - 1
            remaining_time = duration % video.duration
            for _ in range(times_to_loop):
                final_video = concatenate_videoclips([final_video, video])
            final_video = concatenate_videoclips([final_video, video.subclip(0, remaining_time)])
        elif mode == ExtendVideoMode.FREEZE_LAST_FRAME:
            remaining_time = duration - video.duration
            frame = VideoFrameExtractor.get_frame_by_time(video, video.duration)
            frame_freezed_video = generate_video_from_image(frame, remaining_time)
            final_video = concatenate_videoclips([video, frame_freezed_video])
        elif mode in [ExtendVideoMode.SLOW_DOWN, ExtendVideoMode.SPEED_UP_OR_SLOW_DOWN]:
            final_video = ChangeSpeedVideoEffect.apply(final_video, duration)

    return final_video


from manim import *


def base_text(text: str):
    # TODO: Build a preset
    return Text(text)

    self.play(AnimationGroup(
        square.animate.shift(RIGHT),  # Mover a la derecha
        square.animate.rotate(PI / 4)  # Rotar
    ))


from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames.video_frame_handler import VideoFrameHandler
from yta_general_utils.image.parser import ImageParser
from yta_general_utils.programming.enum import YTAEnum as Enum
from moviepy.editor import VideoClip
from typing import Union

import numpy as np


class MoviepyAlphaTransitionProcessingMethod(Enum):
    MEAN = 'mean'
    PURE_BLACK_AND_WHITE = 'pure_black_and_white'

class MoviepyAlphaTransitionHandler:
    """
    Class to handle the functionality related to turning
    alpha clips into moviepy mask clips to be able to 
    apply them and create affects and transitions.
    """
    @staticmethod
    def alpha_clip_to_mask_clip(video: Union[str, VideoClip], method: MoviepyAlphaTransitionProcessingMethod = MoviepyAlphaTransitionProcessingMethod.PURE_BLACK_AND_WHITE):
        """
        The provided alpha 'video' is turned into a moviepy
        mask clip that can be set in any other video by
        using the '.set_mask()' method.

        If you apply the resulting mask to a clip (called 
        'video2' in below) then you can do this below to
        enjoy an incredible transition, or using a black
        background instead of the 'video1' in below.

        CompositeVideoClip([
            video,
            video2
        ]).write_videofile('alpha_masked_transition.mp4')
        """
        video = VideoParser.to_moviepy(video, True, True)
        method = MoviepyAlphaTransitionProcessingMethod.to_enum(method)

        transparent_pixel_in_mask = False
        if video.mask:
            for alpha_frame in video.mask.iter_frames():
                if (alpha_frame < 1.0).any():
                    transparent_pixel_in_mask = True

        # If mask has transparency, we keep the mask
        mask_clip = video.mask
        if not transparent_pixel_in_mask:
            if method == MoviepyAlphaTransitionProcessingMethod.MEAN:
                # Frame by frame, to np.mean and use it as mask
                mask_clip_frames = [VideoFrameHandler.frame_to_mask_frame(frame) for frame in video.iter_frames()]
            elif method == MoviepyAlphaTransitionProcessingMethod.PURE_BLACK_AND_WHITE:
                # Frame by frame, to pure black and white image
                # and turn it into a mask
                mask_clip_frames = [frame_to_mask_frame(frame) for frame in video.iter_frames()]
            
            mask_clip = VideoClip(lambda t: mask_clip_frames[int(t * video.fps)], ismask = True).set_fps(video.fps)

        return mask_clip
    
def frame_to_mask_frame(frame: np.ndarray, do_invert_frame: bool = False):
    """
    Turn the 'frame' parameter into a frame that can
    be used as a mask frame. The provided 'frame'
    should be a frame of a alpha transition clip that
    is almost black and white (or has an alpha layer).

    This method will use an specific strategy to set
    pixels as transparent or opaque by converting the
    frame into a pure black and white image.

    You can use the provided 'frame' but inverted if
    'do_invert_frame' is set as True.
    """
    frame = ImageParser.to_numpy(frame)

    if do_invert_frame:
        frame = VideoFrameHandler.invert_frame(frame)

    frame = VideoFrameHandler.pure_black_and_white_image_to_moviepy_mask_numpy_array(VideoFrameHandler.frame_to_pure_black_and_white_image(frame))

    return frame


from yta_multimedia.video.edition.effect.moviepy.position.objects.base_position_moviepy_effect import BasePositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.move import MoveLinearPositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.static import StayAtPositionMoviepyEffect
from yta_multimedia.video.position import Position
from yta_multimedia.video.edition.effect.moviepy.position.objects.moviepy_slide import MoviepySlide
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, concatenate_videoclips, ColorClip, VideoClip
from typing import Union


class SlideRandomPositionMoviepyEffect(BasePositionMoviepyEffect):
    """
    Effect of appearing from TOP, TOP_LEFT, BOTTOM, RIGHT, etc. 
    staying at the center, and dissapearing from the opposite 
    edge. This animation will spend 1/6 of the time in the 
    entrance, 4/6 of the time staying at the center, and 1/6 of 
    the time in the exit.
    """

    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip]):
        return cls.apply_over_video(video, BasePositionMoviepyEffect.get_black_background_clip(video.duration))

    @classmethod
    def apply_over_video(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], background_video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip]):
        background_video = BasePositionMoviepyEffect.prepare_background_clip(background_video, video)

        random_position = MoviepySlide.get_in_and_out_positions_as_list()

        movement_time = background_video.duration / 6
        stay_time = background_video.duration / 6 * 4

        effect = concatenate_videoclips([
            MoveLinearPositionMoviepyEffect(video.subclip(0, movement_time)).apply_over_video(background_video.subclip(0, movement_time), random_position[0], Position.CENTER),
            StayAtPositionMoviepyEffect(video.subclip(movement_time, movement_time + stay_time)).apply_over_video(background_video.subclip(movement_time, movement_time + stay_time), Position.CENTER),
            MoveLinearPositionMoviepyEffect(video.subclip(movement_time + stay_time, video.duration)).apply_over_video(background_video.subclip(movement_time + stay_time, video.duration), Position.CENTER, random_position[1])
        ])

        return effect


from yta_multimedia.video.frames.numpy_frame_helper import NumpyFrameHelper
from yta_general_utils.image.parser import ImageParser

import numpy as np


WHITE = [255, 255, 255]
BLACK = [0, 0, 0]

class VideoFrameHandler:
    """
    Class to encapsulate functionality related to moviepy
    video frame handling.
    """
    # TODO: This method should be in Image utils
    @staticmethod
    def is_pure_black_and_white_image(image):
        """
        Check if the provided 'image' only contains pure 
        black ([0, 0, 0]) and white ([255, 255, 255]) colors.
        """
        image = ImageParser.to_numpy(image)

        if np.any(~np.all((image == WHITE) | (image == BLACK), axis = -1)):
            return False
        
        return True
    
    @staticmethod
    def pure_black_and_white_image_to_moviepy_mask_numpy_array(image):
        """
        Turn the received 'image' (that must be a pure black
        and white image) to a numpy array that can be used as
        a moviepy mask (by using ImageClip).

        This is useful for static processed images that we 
        want to use as masks, such as frames to decorate our
        videos.
        """
        image = ImageParser.to_numpy(image)

        if not VideoFrameHandler.is_pure_black_and_white_image(image):
            raise Exception(f'The provided "image" parameter "{str(image)}" is not a black and white image.')

        # Image to a numpy parseable as moviepy mask
        mask = np.zeros(image.shape[:2], dtype = int) # 3col to 1col
        mask[np.all(image == WHITE, axis = -1)] = 1 # white to 1 value

        return mask
    
    def frame_to_pure_black_and_white_image(frame):
        """
        Process the provided moviepy clip mask frame (that
        must have values between 0.0 and 1.0) or normal clip
        frame (that must have values between 0 and 255) and
        convert it into a pure black and white image (an
        image that contains those 2 colors only).

        This method returns a not normalized numpy array of only
        2 colors (pure white [255, 255, 255] and pure black
        [0, 0, 0]), perfect to turn into a mask for moviepy clips.

        This is useful when handling an alpha transition video 
        that can include (or not) an alpha layer but it is also
        clearly black and white so you transform it into a mask
        to be applied on a video clip.
        """
        frame = ImageParser.to_numpy(frame)

        if not VideoFrameHandler.frame_is_mask_clip_frame(frame) and not VideoFrameHandler.frame_is_normal_clip_frame(frame):
            raise Exception('The provided "frame" parameter is not a moviepy mask clip frame nor a normal clip frame.')
        
        if VideoFrameHandler.frame_is_normal_clip_frame(frame):
            # TODO: Process it with some threshold to turn it
            # into pure black and white image (only those 2
            # colors) to be able to transform them into a mask.
            threshold = 220
            white_pixels = np.all(frame >= threshold, axis = -1)

            # Image to completely and pure black
            new_frame = np.array(frame)
            
            # White pixels to pure white
            new_frame[white_pixels] = [255, 255, 255]
            new_frame[~white_pixels] = [0, 0, 0]
        elif VideoFrameHandler.frame_is_mask_clip_frame(frame):
            transparent_pixels = frame == 1

            new_frame = np.array(frame)
            
            # Transparent pixels to pure white
            new_frame[transparent_pixels] = [255, 255, 255]
            new_frame[~transparent_pixels] = [0, 0, 0]

        return new_frame
    
    @staticmethod
    def frame_is_normal_clip_frame(frame: np.ndarray):
        """
        Checks if the provided 'frame' numpy array is recognized as
        a frame of a normal moviepy clip with values between 0 and
        255.

        This numpy array should represent a frame of a clip.
        
        A non-modified clip is '.ndim = 3' and '.dtype = np.uint8'.
        """
        return NumpyFrameHelper.is_rgb_not_normalized(frame)
        
    @staticmethod
    def frame_is_mask_clip_frame(frame: np.ndarray):
        """
        Checks if the provided 'mask_clip' numpy array is recognized
        as an original moviepy mask clip with values between 0 and 1.
        This numpy array should represent a frame of a mask clip.
        
        A non-modified mask clip is '.ndim = 2' and '.dtype = np.float64'.
        """
        return NumpyFrameHelper.is_alpha_normalized(frame)
        
    @staticmethod
    def invert_frame(frame: np.ndarray):
        """
        Invert the values of the provided 'frame', that can be
        a moviepy normal clip frame (with values between 0 and
        255) or a mask clip frame (with values between 0 and 1).

        This method invert the values by applying the max value
        (255 for normal frame, 1 for mask frame) minus each value
        in the numpy array.

        This method returns the numpy array inverted.
        """
        if not VideoFrameHandler.frame_is_normal_clip_frame(frame) and not VideoFrameHandler.frame_is_mask_clip_frame(frame):
            raise Exception('The provided "frame" is not actually a moviepy normal clip frame nor a mask clip frame.')
        
        if VideoFrameHandler.frame_is_normal_clip_frame(frame):
            frame = 255 - frame
        elif VideoFrameHandler.frame_is_mask_clip_frame(frame):
            frame = 1 - frame

        return frame
    
    @staticmethod
    def frame_to_mask_frame(frame: np.ndarray):
        """
        Turn the provided 'frame', that must be represented by
        RGB pixel values ([0-255, 0-255, 0-255]), to a mask
        frame represented by a single value between 0 and 1
        ([0.0-1.0]).
        """
        frame = ImageParser.to_numpy(frame)

        if not VideoFrameHandler.frame_is_normal_clip_frame(frame):
            raise Exception('The provided "frame" is not actually a moviepy normal clip frame.')
        
        frame = np.mean(frame, axis = -1) / 255.0

        return frame
    


from yta_general_utils.programming.env import get_current_project_env

import google.generativeai as genai


class GeminiAI:
    """
    Class that represents the Google Gemini AI generative model
    to work with it.
    """
    MODEL_NAME = 'gemini-1.5-flash'

    def __init__(self):
        genai.configure(api_key = get_current_project_env('GEMINI_API_KEY'))
        self.model = genai.GenerativeModel(self.MODEL_NAME)

    def ask(self, prompt: str):
        """
        Asks Gemini AI (gemini-1.5-flash) model by using the provided prompt, waits for the response
        and returns it.
        """
        chat = self.model.start_chat()
        response = chat.send_message(
            prompt,
        )

        return response.text


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, clips_array, VideoClip, ColorClip
from typing import Union
from math import log as math_log, pow as math_pow


class MultipliedVideoEffect(VideoEffect):
    """
    Generates a clips array with the provided 'clip' being shown
    'times' times (this parameter must be a pow of 4). This
    method has been created to be used internally with our own
    default methods.
    """
    
    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], times: int = None):
        """
        Applies the effect to the provided 'video'.
        """
        video = VideoEffect.parse_moviepy_video(video)

        if not times:
            times = 4

        audio = video.audio
        size = (video.w, video.h)

        # We will dynamically build the matrix
        row = []
        group = []
        # 4^3 = 64 => 8x8 = 2^3x2^3  and  4^2 = 16 => 4x4 = 2^2x2^2
        range_limit_value = math_pow(2, math_log(times, 4))
        for i in range(int(times / range_limit_value)):
            row = []
            for j in range(int(times / range_limit_value)):
                row.append(video)
            group.append(row)

        # When building 'clips_array' you sum the resolutions, so if you add four videos
        # of 1920x1080, you'll get one video of 4x(1920x1080) that will be impossible to
        # exported and unexpected. We resize it to avoid this and we don't resize each
        # clip before because they lose quality
        return clips_array(group).resize(size).set_audio(audio)


# https://open.spotify.com/get_access_token   Here you get access token to spotify (for what?)
from yta_general_utils.programming.env import get_current_project_env
from musixmatch import Musixmatch

import requests


MUSIXMATCH_API_KEY = get_current_project_env('MUSIXMATCH_API_KEY')

def get_lyrics_from_musixmatch(song_id):
    musixmatch = Musixmatch(MUSIXMATCH_API_KEY)

    # example id: 15953433
    return musixmatch.track_lyrics_get(song_id)

def get_song_lyrics_srt(author, song, output_filename):
    """
    Looks for the 'author's 'song' lyrics and, if found, builds an SRT file
    with those lyrics and stores it locally as 'output_filename'.
    """
    lyrics = get_timestamped_lyrics(author, song)

    if lyrics:
        timestamped_lyrics_to_srt(lyrics, output_filename)

def get_timestamped_lyrics(author, song):
    # Found here: https://stackoverflow.com/questions/62253539/api-with-lyrics-and-timestamps
    # This below works
    #url = 'https://api.textyl.co/api/lyrics?q=alan%20walker%20faded'
    # TODO: Find a better solution because this one doesn't take care about when the
    # author is not singing. The lyric is shown even when voice silence, and I don't
    # want that. This returns the moment in which a sentence starts, but not when it
    # ends, just when the next one starts, that is not the same.
    url = 'https://api.textyl.co/api/lyrics?q=' + author + ' ' + song

    response = requests.get(url)
    response = response.json()

    return response

def timestamped_lyrics_to_srt(timestamped_lyrics, output_filename):
    index = 0
    srt_content = ''
    while index < (len(timestamped_lyrics) - 1):
        # TODO: This fails if only 1 timestamp_lyric element
        timestamped_lyric = timestamped_lyrics[index]
        srt_content += str(index + 1) + '\n'

        m, s = divmod(timestamped_lyric['seconds'], 60)
        h, m = divmod(m, 60)
        
        srt_content += '{:d}:{:02d}:{:02d},000'.format(h, m, s) + ' --> '

        m, s = divmod(timestamped_lyrics[index + 1]['seconds'], 60)
        h, m = divmod(m, 60)
        srt_content += '{:d}:{:02d}:{:02d},000'.format(h, m, s) + '\n'

        srt_content += timestamped_lyric['lyrics'] + '\n'
        srt_content += '\n'

        index += 1

    # Write .srt file
    with open(output_filename, 'w') as outfile:
        outfile.write(srt_content)

       
    """
    # This is the SRT format
    1
    00:00:00,000 --> 00:00:02,500
    Welcome to the Example Subtitle File!

    """





from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames.video_frame_handler import VideoFrameHandler
from moviepy.editor import VideoClip, VideoClip
from typing import Union


class VideoHandler:
    """
    Class created to simplify and encapsulate the working process
    with moviepy videos.
    """
    @staticmethod
    def invert_video(video: Union[str, VideoClip]):
        """
        Invert the received 'video' (that must be a moviepy 
        mask or normal clip) and return it inverted as a
        VideoClip. If the provided 'video' is a mask, this 
        will be also a mask.

        If the 'clip' provided is a mask clip, remember to
        set it as the new mask of your main clip.

        This inversion is a process in which the numpy array
        values of each frame are inverted by substracting the
        highest value. If the frame is an RGB frame with 
        values between 0 and 255, it will be inverted by 
        doing 255 - X on each frame pixel value. If it is
        normalized and values are between 0 and 1 (it is a 
        mask clip frame), by doing 1 - X on each mask frame
        pixel value.
        """
        video = VideoParser.to_moviepy(video)

        mask_frames = [VideoFrameHandler.invert_frame(frame) for frame in video.iter_frames()]

        return VideoClip(lambda t: mask_frames[int(t * video.fps)], ismask = video.ismask).set_fps(video.fps)


from yta_general_utils.file.checker import FileValidator
from yta_multimedia.video.audio import set_audio_in_video
from moviepy.editor import VideoFileClip
from numba import njit

import pygame as pg
import cv2


@njit(fastmath=True)
def accelerate_conversion(image, width, height, ascii_coeff, step):
    array_of_values = []
    for x in range(0, width, step):
        for y in range(0, height, step):
            char_index = image[x, y] // ascii_coeff
            if char_index:
                array_of_values.append((char_index, (x, y)))
    return array_of_values

class AsciiFilteredVideo:
    def __init__(self, input_filename, output_filename, font_size = 12):
        pg.init()
        self.path = input_filename
        self.capture = cv2.VideoCapture(input_filename)
        self.image = self.get_image()
        self.RES = self.WIDTH, self.HEIGHT = self.image.shape[0], self.image.shape[1]
        self.surface = pg.display.set_mode(self.RES)
        self.clock = pg.time.Clock()

        self.ASCII_CHARS = ' .",:;!~+-xmo*#W&8@'
        self.ASCII_COEFF = 255 // (len(self.ASCII_CHARS) - 1)

        self.font = pg.font.SysFont('Courier', font_size, bold = True)
        self.CHAR_STEP = int(font_size * 0.6)
        self.RENDERED_ASCII_CHARS = [self.font.render(char, False, 'white') for char in self.ASCII_CHARS]

        self.rec_fps = self.capture.get(cv2.CAP_PROP_FPS)
        self.record = True
        self.recorder = cv2.VideoWriter(output_filename, cv2.VideoWriter_fourcc(*'mp4v'), self.rec_fps, self.RES)

    def get_frame(self):
        frame = pg.surfarray.array3d(self.surface)
        return cv2.transpose(frame)

    def record_frame(self):
        if self.record:
            frame = self.get_frame()
            self.recorder.write(frame)
            #cv2.imshow('Frame', frame)
            if cv2.waitKey(1) & 0xFF == 27:
                self.record = not self.record
                cv2.destroyAllWindows()

    def draw_converted_image(self):
        self.image = self.get_image()
        array_of_values = accelerate_conversion(self.image, self.WIDTH, self.HEIGHT, self.ASCII_COEFF, self.CHAR_STEP)
        for char_index, pos in array_of_values:
            self.surface.blit(self.RENDERED_ASCII_CHARS[char_index], pos)

    def get_image(self):
        # self.cv2_image = cv2.imread(self.path)
        ret, self.cv2_image = self.capture.read()
        if not ret:
            exit()
        transposed_image = cv2.transpose(self.cv2_image)
        gray_image = cv2.cvtColor(transposed_image, cv2.COLOR_BGR2GRAY)
        return gray_image

    def draw_cv2_image(self):
        resized_cv2_image = cv2.resize(self.cv2_image, (640, 360), interpolation=cv2.INTER_AREA)
        cv2.imshow('img', resized_cv2_image)

    def draw(self):
        self.surface.fill('black')
        self.draw_converted_image()
        # self.draw_cv2_image()

    def save_image(self):
        pygame_image = pg.surfarray.array3d(self.surface)
        cv2_img = cv2.transpose(pygame_image)
        cv2.imwrite('ascii_image.jpg', cv2_img)

    def run(self):
        while True:
            for i in pg.event.get():
                if i.type == pg.QUIT:
                    exit()
                elif i.type == pg.KEYDOWN:
                    if i.key == pg.K_s:
                        self.save_image()
                    if i.key == pg.K_r:
                        self.record = not self.record
            self.record_frame()
            self.draw()
            pg.display.set_caption(str(self.clock.get_fps()))
            pg.display.flip()
            self.clock.tick()

def to_ascii(video_filename: str, output_filename: str, kwargs):
    """
    Gets the provided 'input_filename' video, turns it into ascii
    video and stores the result in a new video as 'output_filename'.

    Optional parameter is 'font_size' (default value is 12).
    """
    if not video_filename:
        return None
    
    if not FileValidator.file_is_video_file(video_filename):
        return None
    
    if not output_filename:
        return None
    
    font_size = kwargs.get('font_size', 12)

    app = AsciiFilteredVideo(video_filename, output_filename, font_size)
    app.run()

    output_videoclip = VideoFileClip(video_filename)
    output_videoclip_with_sound = set_audio_in_video(output_filename, output_videoclip.audio)
    output_videoclip_with_sound.write_videofile(output_filename)




from yta_multimedia.video.position import Position


class MoviepyPosition:
    """
    Enum class to encapsulate and simplify the way we work with
    positions in the moviepy video scene system. This is just a
    shortcut to using directly a Position Enum class and its
    moviepy methods.
    """
    @staticmethod
    def get_position(position: Position, video, background_video):
        """
        This method will calculate the (x, y) tuple position for the provided
        'video' over the also provided 'background_video' that would be,
        hypothetically, a 1920x1080 black color background static image. The
        provided 'position' will be transformed into the (x, y) tuple according
        to our own definitions.
        """
        position = Position.to_enum(position)

        return position.get_moviepy_position(video, background_video)

# TODO: Import all effects to have them here available (?)


from yta_multimedia.video.frames.video_frame_handler import VideoFrameHandler
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.file.enums import FileType
from yta_general_utils.file.filename import filename_is_type
from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy.editor import VideoFileClip, CompositeVideoClip, ColorClip, ImageClip, ImageSequenceClip, VideoClip
from typing import Union


class VideoParser:
    """
    Class to simplify the way we parse video parameters.
    """
    @classmethod
    def to_moviepy(cls, video: Union[str, VideoFileClip, CompositeVideoClip, ColorClip, ImageClip], do_include_mask: bool = False, do_check_duration: bool = False):
        """
        This method is a helper to turn the provided 'video' to a moviepy
        video type. If it is any of the moviepy video types specified in
        method declaration, it will be returned like that. If not, it will
        be load as a VideoFileClip if possible, or will raise an Exception
        if not.

        The 'do_include_mask' parameter includes the mask in the video if
        True value provided. The 'do_check_duration' parameter checks and
        updates the real video duration to fix a bug in moviepy lib.
        """
        if not video:
            raise Exception('No "video" provided.')
        
        # TODO: Maybe check if subclass of VideoClip
        if not PythonValidator.is_string(video) and not PythonValidator.is_instance(video, VideoFileClip) and not PythonValidator.is_instance(video, CompositeVideoClip) and not PythonValidator.is_instance(video, ColorClip) and not PythonValidator.is_instance(video, ImageClip) and not PythonValidator.is_instance(video, ImageSequenceClip) and not PythonValidator.is_instance(video, VideoClip):
            raise Exception('The "video" parameter provided is not a valid type. Check valid types in method declaration.')
        
        if PythonValidator.is_string(video):
            if not filename_is_type(video, FileType.VIDEO):
                raise Exception('The "video" parameter provided is not a valid video filename.')
            
            if not FileValidator.file_is_video_file(video):
                raise Exception('The "video" parameter is not a valid video file.')
            
            video = VideoFileClip(video, has_mask = do_include_mask)

        if do_check_duration:
            # We need to fix an existing bug in moviepy lib
            # see https://github.com/Zulko/moviepy/issues/1826
            video = verify_and_update_duration(video)

        # TODO: This below just adds a mask attribute but
        # without fps and empty, so it doesn't make sense
        # if do_include_mask and not video.mask:
        #     video = video.add_mask()

        return video
    
    @staticmethod
    def is_normal_clip(clip: Union[str, VideoClip]):
        """
        Normal clips have frames as numpy arrays with values between
        0 and 255 (where 255 means full of color).

        A non-modified clip is '.ndim = 3' and '.dtype = uint8'.
        """
        clip = VideoParser.to_moviepy(clip)

        # TODO: Should I do 'is_subclass' instead of 'is_instance' (?)
        # TODO: I'm not sure, maybe a CompositeVideoClip is a mask (?)
        if not PythonValidator.is_instance(clip, VideoClip):
            # TODO: False or Exception (?)
            raise Exception('The provided "clip" parameter is not a moviepy VideoClip instance.')
        
        return VideoFrameHandler.frame_is_normal_clip_frame(clip.get_frame(t = 0))

    @staticmethod
    def is_mask_clip(clip: Union[str, VideoClip]):
        """
        Mask clips have frames as numpy arrays with values only between
        0.0 and 1.0 (where 1.0 means completely opaque). 
        
        A non-modified mask is '.ndim = 2' and '.dtype = float64'.
        """
        clip = VideoParser.to_moviepy(clip)

        # TODO: Should I do 'is_subclass' instead of 'is_instance' (?)
        # TODO: I'm not sure, maybe a CompositeVideoClip is a mask (?)
        if not PythonValidator.is_instance(clip, VideoClip):
            # TODO: False or Exception (?)
            raise Exception('The provided "clip" parameter is not a moviepy VideoClip instance.')

        return VideoFrameHandler.frame_is_mask_clip_frame(clip.get_frame(t = 0))

def verify_and_update_duration(video):
    """
    Try to subclip the provided 'video' with the last
    frames detected by moviepy and check if the duration
    is the real one and updates it if not. This method
    returns the video updated to the its new duration.

    Moviepy has a bug in which some videos are detected
    larger than they actually are, and that makes the 
    software fails when trying to work with all those 
    wrong detected videos.

    When trying to subclip a moviepy video, the system
    will fail when the 't_start' parameter is larger or
    equal to the actual video duration (not the one
    detected by moviepy), so we are using that slow
    function to detect the real duration and updating
    the video accordingly.
    """
    frame_time = 1 / video.fps
    
    for frame_number in reversed(range(int(video.fps * video.duration))):
        try:
            real_duration = frame_number * frame_time
            video.subclip(real_duration, 9999)
            # 't_start' can't be the last frame but the penultimate
            # so the real duration is one 'frame_time' later
            real_duration += frame_time
            break
        except:
            pass

    video = video.subclip(0, real_duration)

    return video


from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.constants import SCENE_HEIGHT, SCENE_WIDTH, HALF_SCENE_HEIGHT
from manim import *


class CinematicBarsManimAnimation(BaseManimAnimation):
    """
    Black bars that appear to make the scene cinematic.
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, duration: float = 2, output_filename: str = 'output.mov'):
        """
        Checks and validates the provided parameters and generates
        the manim animation if those parameters are valid.
        """
        # Check and validate all parameters
        parameters = {}

        if super().parameter_is_mandatory('duration', self.required_parameters) and not duration:
            raise Exception('Field "duration" is mandatory. Aborting manim creation...')
        if duration < 0 or duration > 100:
            raise Exception('Field "duration" value is not valid. Must be between 0 and 100')
        
        parameters['duration'] = duration

        if not output_filename:
            output_filename = 'output.mov'

        # Generate the animation when parameters are valid
        super().generate(parameters, output_filename = output_filename)

        return output_filename

    def animate(self):
        top_bar = Rectangle(BLACK, stroke_width = 0, fill_color = BLACK, fill_opacity = 1, height = SCENE_HEIGHT * 0.15, width = SCENE_WIDTH)
        bottom_bar = top_bar.copy()

        START_TOP_POSITION = (0, HALF_SCENE_HEIGHT + (top_bar.height / 2), 0)
        STAY_TOP_POSITION = (0, HALF_SCENE_HEIGHT, 0)
        START_BOTTOM_POSITION = (0, -HALF_SCENE_HEIGHT - (bottom_bar.height / 2), 0)
        STAY_BOTTOM_POSITION = (0, -HALF_SCENE_HEIGHT, 0)

        top_bar.move_to(START_TOP_POSITION)
        bottom_bar.move_to(START_BOTTOM_POSITION)

        self.add(top_bar)
        self.add(bottom_bar)

        # Appearing and dissapearing animation can be 2 seconds
        # longer as maximum (each one). If this happens, the 
        # stay will be longer
        animation_duration = 0.2 * self.parameters['duration']
        stay_duration = 0.6 * self.parameters['duration']
        if animation_duration > 2:
            stay_duration += (animation_duration - 2) * 2

        self.play(AnimationGroup([
            top_bar.animate.move_to(STAY_TOP_POSITION),
            bottom_bar.animate.move_to(STAY_BOTTOM_POSITION)
        ]), run_time = animation_duration, rate_func = linear)

        self.wait(stay_duration)

        self.play(AnimationGroup([
            top_bar.animate.move_to(START_TOP_POSITION),
            bottom_bar.animate.move_to(START_BOTTOM_POSITION)
        ]), run_time = animation_duration, rate_func = linear)


from yta_multimedia.video.edition.effect.moviepy.position.interpolation import NormalizedCoordinate
from yta_general_utils.math import Math


class CoordinateCenter:
    """
    Class that encapsulates a coordinate (x, y) representing the
    center of a video in a movement effect envinronment. This 
    class is used as type to represent the position in which we
    want to place a video, considering this position as the
    center of the video.
    """
    position = (0, 0)
    """
    The (x, y) tuple containing the position coordinate.
    """

    def __init__(self, x: int, y: int):
        if not x and x != 0:
            raise Exception('No "x" provided.')
        
        if not y and y != 0:
            raise Exception('No "y" provided.')
        
        # TODO: Check 'x' and 'y' are numbers and cast to int
        
        self.position = (x, y)

    def get_x(self):
        return self.position[0]
    
    def get_y(self):
        return self.position[1]
    
    def to_tuple(self):
        return self.position
    
    def to_array(self):
        return [self.position[0], self.position[1]]
        
    def to_normalized_coordinate(self, min_value = -10000, max_value = 10000) -> NormalizedCoordinate:
        """
        Normalizes the coordinate to turn it into a value between
        0 and 1 to be able to use it in our GraphRateFunction 
        system. As we are working in a coordinates system that
        involves 1920x1080 scenarios and our videos can be 
        completely outside of the scene, I use this default values.
        """
        return NormalizedCoordinate(Math.normalize(self.position[0], min_value, max_value), Math.normalize(self.position[1], min_value, max_value))
    
    def recalculate_for_video(self, video, background_video):
        """
        This method will return the coords (x, y) in which we need to place the
        'video' to have its center in the desired ('x', 'y') position over the 
        also provided 'background_video' by making some calculations as below.

        Imagine a scene of a 1920x1080 black background and that the 'x' and 'y'
        you give as parameters are the center of the 'video'. We will calculate 
        to place the provided 'video' there in the real situation, over the 
        'background_video' that could be not 1920x1080.
        """
        # TODO: Implement checkings

        # Considering a 1920x1080 scene, recalculate actual coords
        x = (int) (background_video.w * self.position[0] / 1920)
        y = (int) (background_video.h * self.position[1] / 1080)

        # Coordinates were from center, so adapt to upper left corner
        x -= (video.w / 2)
        y -= (video.h / 2)

        x = int(x)
        y = int(y)

        return (x, y)